# -*- coding: utf-8 -*-
import click

from api_client.runs_client import RunsClient
from api_client.workspaces_client import WorkspacesClient
from spell_cli.exceptions import api_client_exception_handler
from spell_cli.utils import (
    prettify_time,
    tabulate_rows,
    HiddenOption,
    truncate_string,
)
from spell_cli.log import logger


# display order of columns
COLUMNS = [
    "id",
    "ws_name",
    "command",
    "git_commit_hash",
    "pretty_status",
    "gpu",
]

# title lookup
TITLES = {
    "id": "ID",
    "ws_name": "WORKSPACE",
    "command": "COMMAND",
    "git_commit_hash": "COMMIT",
    "pretty_status": "STATUS",
    "gpu": "MACHINE TYPE",
}

status_names = {
    "user_requested": "Requested",
    "machine_requested": "Queued",
}


COMMIT_DISPLAY_LENGTH = 8
COMMAND_DISPLAY_LENGTH = 30


# prettify_status returns a nice human readable status for display to the user
def prettify_status(run):
    status = status_names.get(run.status, run.status).title()
    if run.user_exit_code is not None:
        status += " ({})".format(run.user_exit_code)
    if run.started_at is not None and run.ended_at is None:
        status += " [{}]".format(prettify_time(run.started_at, elapsed=True))
    elif run.ended_at is not None:
        status += " -- {}".format(prettify_time(run.ended_at))
    # TODO(libeks): Remove `requested` once fully migrated
    if run.status in ("requested", "user_requested", "machine_requested"):
        return click.style(status + " -- {}".format(prettify_time(run.created_at)), fg='yellow')
    elif run.status in ("building", "mounting", "running", "saving", "pushing"):
        return click.style("*"+status, fg='green')
    elif run.status in ("failed", "complete", "killing", "killed", "stopping", "stopped"):
        return status


@click.command(name="ps",
               short_help="Display run statuses")
@click.option("--raw", is_flag=True,
              help="Display output in raw format", cls=HiddenOption)
@click.option("--trunc/--no-trunc", default=True,
              help="Truncate command output")
@click.option("-w", "--workspace", default=None,
              help="Workspace name to filter runs by")
@click.pass_context
def ps(ctx, raw, trunc, workspace):
    token = ctx.obj["config_handler"].config.token

    # retrieve runs from API
    with api_client_exception_handler():
        logger.info("Retrieving runs from Spell")
        run_client = RunsClient(token=token, **ctx.obj["client_args"])
        if workspace is None:
            runs = run_client.list_runs()
        else:
            workspace_client = WorkspacesClient(token=token, **ctx.obj["client_args"])
            workspace = workspace_client.get_workspace_by_name(workspace)
            runs = run_client.list_runs(str(workspace.id))

    # prettify attrs for output
    for run in runs:
        run.id = str(run.id)
        if trunc:
            run.command = truncate_string(run.command, COMMAND_DISPLAY_LENGTH)
        run.ws_name = run.workspace.name
        run.git_commit_hash = run.git_commit_hash[:COMMIT_DISPLAY_LENGTH]
        run.pretty_status = prettify_status(run)
        run.pretty_created_at = prettify_time(run.created_at)

    tabulate_rows(runs,
                  headers=[TITLES[col] for col in COLUMNS],
                  columns=COLUMNS,
                  raw=raw)
